package cds.stage;

import cds.aladin.Calib;
import cds.aladin.Coord;

/**
 * Gestion des projections clestes
 * @author Pierre Fernique [CDS]
 * @version 1.0 cration - avril 2011
 */
public class Proj {
   private Calib calib;
   private double zoom;
   private double xc, yc;
   private double ra,dec;
   private double rota;
   private int PROJECTION = Calib.SIN;
   
   /**
    * Cration d'une projection cleste
    * @param xc Abscisse du centre de la projection
    * @param yc Ordonne du centre de la projection
    * @param zoom Facteur de zoom
    */
   public Proj(double xc,double yc,double zoom) {
      ra=0;
      dec=0;
      rota=0;
      calib = new Calib(ra,dec,0,0,500,500, 60*4,60*4,rota,PROJECTION,false,Calib.FK5);
      this.xc=xc;
      this.yc=yc;
      this.zoom=zoom;
   }
   
   /** Modification du facteur de zoom */
   public void setZoom(double zoom) { this.zoom=zoom; }
   
   /** Fournit le facteur de zoom courant */
   public double getZoom() { return zoom; }
   
   /** Modification de la position du centre de projection (en XY)
    * @param xc  Abscisse
    * @param yc Ordonne
    */
   public void setXYCenter(double xc,double yc) { this.xc=xc; this.yc=yc; }
   
   /** Fournit la position du centre de projection (en XY)
    * @return l'absisse et l'ordonne
    */
   public double [] getXYCenter() { return new double[] { xc,yc }; }
   
   /** Dcalage de la coordonne cleste corresondante au centre de projection
    * => en projection SINUS, entrainera une rotation
    * @param deltaRa dcalage en ascension droite
    * @param deltaDec dcalage en dclinaison
    */
   protected void deltaProjCenter(double deltaRa,double deltaDec) {
      ra += deltaRa;
      dec += deltaDec;
      if( dec>89.95 ) { dec=180-dec; ra+=180; rota+=180; }
      else if( dec<-89.95 ) { dec=-dec-180; ra+=180; rota-=180; }
      if( ra>360 ) ra-=360;
      else if( ra<0 ) ra+=360;
      if( rota>360 ) rota-=360;
      else if( rota<0 ) rota+=360;
      calib = new Calib(ra,dec,0,0,500,500, 60*4,60*4,rota,PROJECTION,false,Calib.FK5);
   }
   
   /** Pour une position (coo.x,coo.y), calcule la coordonne cleste (coo.al,coo.del)
    * Si problme, la coordonne cleste sera (Double.NaN,Double.NaN)
    * @param coo les coordonnes (x,y en entre -  al,del en sortie)
    */
   public void computeCoord(Coord coo) {
      if( Double.isNaN(coo.x) ) { coo.al=Double.NaN; coo.del=Double.NaN; return; }
      Coord c = new Coord();
      c.x = (coo.x-xc)/zoom;
      c.y =( coo.y-yc)/zoom;
      try {
         calib.GetCoord(c);
         coo.al=c.al;
         coo.del=c.del;
      } catch( Exception e ) { coo.al=Double.NaN; coo.del=Double.NaN; }
   }

   /** Pour une coordonne cleste (coo.al,coo.del) calcule la position (coo.x,coo.y)
    * Si problme, la position retourne sera (Double.NaN,Double.NaN)
    * @param coo les coordonnes ( al,del en entre -  x,y en sortie)
    */
   public void computeXY(Coord coo) {
      if( Double.isNaN(coo.al) ) { coo.x=Double.NaN; coo.y=Double.NaN; return; }
      Coord c = new Coord(coo.al,coo.del);
      try {
         calib.GetXY(c);
         coo.x = (c.x*zoom)+xc;
         coo.y = (c.y*zoom)+yc;
      } catch( Exception e ) { coo.x=Double.NaN; coo.y=Double.NaN; }
   }
}
